"""
    Main GUI.
"""
from datetime import datetime

import requests
import xbmc
import xbmcplugin

from resources.lib.api.api import API
from resources.lib.api.open_subtitles import OpenSubtitles
from resources.lib.const import SETTINGS, ROUTE, SERVICE_EVENT, LANG, \
    SERVICE, STRINGS, GENERAL, COMMAND, HTTP_METHOD, COUNT, MEDIA_TYPE, MEDIA_SERVICE, DB_TABLE, languages, RENDER_TYPE, \
    DIR_TYPE, CONTEXT_MENU
from resources.lib.gui import InfoDialog, InfoDialogType, MediaItem
from resources.lib.gui.renderers.dialog_renderer import DialogRenderer
from resources.lib.gui.renderers.directory_renderer import DirectoryRenderer
from resources.lib.gui.renderers.media_info_renderer import MediaInfoRenderer
from resources.lib.gui.renderers.media_list_renderer import MediaListRenderer
from resources.lib.kodilogging import logger
from resources.lib.routing.router import router
from resources.lib.storage.settings import settings
from resources.lib.storage.sqlite import SQLiteStorage
from resources.lib.utils import streams
from resources.lib.utils.csfd_tips import get_csfd_tips
from resources.lib.utils.kodiutils import get_string, time_limit_expired, show_settings, hash_password, \
    kodi_json_request, get_log_file_path, read_log, clean_log, post_log, show_log_dialog
from resources.lib.utils.url import Url
from resources.lib.utils.addtolib import add_to_library


class StreamCinema:
    def __init__(self, provider, api):
        self._api = api
        self._provider = provider
        self.db = {
            DB_TABLE.WATCH_HISTORY: SQLiteStorage.WatchHistory()
        }

        self.directory_renderer = DirectoryRenderer()
        self.media_renderer = MediaListRenderer()

        self.command_map = {
            COMMAND.OPEN_SETTINGS: self.show_settings,
            COMMAND.CLEAR_PATH: self.clear_path,
            COMMAND.SHOW_LOG_PATH: self.show_log_path,
            COMMAND.UPLOAD_LOG: self.upload_log,
            COMMAND.SET_PROVIDER_CREDENTIALS: self.set_provider_credentials,
            COMMAND.SET_SUBTITLES_CREDENTIALS: self.set_subtitles_credentials,
            COMMAND.REFRESH_PROVIDER_TOKEN: self.refresh_provider_token,
            COMMAND.ADD_TO_LIBRARY: add_to_library,
            COMMAND.DELETE_HISTORY_ALL: self.delete_history_all,
        }

        self.route_map = {
            ROUTE.ROOT: self.main_menu,
            ROUTE.SEARCH: self.search,
            ROUTE.GET_MEDIA: self.get_media,
            ROUTE.MOVIES: self.movies,
            ROUTE.TV_SHOWS: self.tv_shows,
            ROUTE.COMMAND: self.command,
            ROUTE.A_TO_Z: self.a_to_z,
            ROUTE.GENRE_MENU: self.genre_menu,
            ROUTE.GET_STREAMS: self.get_streams,
            ROUTE.STUDIO_MENU: self.studio_menu,
            ROUTE.CSFD_TIPS: self.csfd_tips,
            ROUTE.WATCHED: self.watched,
            ROUTE.DELETE_HISTORY_ITEM: self.delete_history_item
        }
        self.permission_map = {
            True: self.play_stream,
            False: self.login_play_stream
        }

        self.dir_type_render = {
            RENDER_TYPE.DEFAULT: self.get_media_default,
            RENDER_TYPE.A_Z: self.get_media_az,
            RENDER_TYPE.SEARCH + '?': self.get_media_search
        }

        self.context_menu_map = {
            ROUTE.ROOT: [],
            ROUTE.WATCHED: [CONTEXT_MENU.DELETE_HISTORY],
        }

    def dispatch(self, handle, url, query):
        self.route_map[url](handle, **dict(Url.parse_qsl(query.lstrip('?'))))

    def command(self, handle, command):
        xbmcplugin.endOfDirectory(handle)
        self.command_map[command](handle)

    def main_menu(self, handle):
        self.directory_renderer.main_menu(handle)

    def movies(self, handle):
        self.directory_renderer.movies(handle)

    def tv_shows(self, handle):
        self.directory_renderer.tv_shows(handle)

    def genre_menu(self, handle, media_type=None):
        response = self.api_request(HTTP_METHOD.GET, self._api.FILTER.count(media_type, COUNT.GENRES))
        self.directory_renderer.genre_menu(handle, media_type, response.get('data'))

    def studio_menu(self, handle, media_type=None):
        response = self.api_request(HTTP_METHOD.GET, self._api.FILTER.count(media_type, COUNT.STUDIOS))
        self.directory_renderer.studio_menu(handle, media_type, response.get('data'))

    def search(self, handle, media_type=MEDIA_TYPE.ALL):
        self.directory_renderer.search(handle, media_type)

    def a_to_z(self, handle, media_type=None, letter=''):
        response = self.api_request(HTTP_METHOD.GET, self._api.FILTER.az_count(media_type, letter))
        self.directory_renderer.a_to_z_menu(handle, media_type, letter, response)

    def get_media(self, handle, url, render_type, media_type):
        self.dir_type_render[render_type](handle, url, render_type, media_type)

    def get_media_default(self, handle, url, render_type, media_type):
        response = self.api_request(HTTP_METHOD.GET, url)
        self.render_media(handle, response, render_type=render_type, media_type=media_type)

    def get_media_search(self, handle, url, render_type, dir_type):
        # TODO: Remove this dirty hack but I have no ducking idea why there is question mark in search URL ?????
        response = self.api_request(HTTP_METHOD.GET, url)
        self.render_media(handle, response, MediaInfoRenderer.TITLE.mixed, render_type, dir_type)

    def get_media_az(self, handle, url, render_type, media_type, context_menu=ROUTE.ROOT):
        query = Url.parse_qs(url)
        letters = query.get('value')[0]
        media_list = self.api_request(HTTP_METHOD.GET, url)
        media_list_gui = self.media_renderer.build_media_list(media_list, RENDER_TYPE.DEFAULT, self.context_menu_map[context_menu], MediaInfoRenderer.TITLE.a_z, letters)
        media_list_gui.sort(key=lambda x: x.title_raw)
        self._render_call(handle, media_list, media_list_gui, render_type, media_type)

    def render_media(self, handle, media_list, title_renderer=MediaInfoRenderer.TITLE.default,
                     render_type=RENDER_TYPE.DEFAULT, media_type=MEDIA_TYPE.ALL, context_menu=ROUTE.ROOT):
        media_list_gui = self.media_renderer.build_media_list(media_list, RENDER_TYPE.DEFAULT, self.context_menu_map[context_menu], title_renderer)
        self._render_call(handle, media_list, media_list_gui, render_type, media_type)

    def _render_call(self, handle, media_list, media_list_gui, render_type=RENDER_TYPE.DEFAULT, media_type=MEDIA_TYPE.ALL):
        self.media_renderer(handle, media_list_gui, media_list.get('pagination'), render_type, media_type)

    def csfd_tips(self, handle):
        url = API.FILTER.service(MEDIA_TYPE.MOVIE, MEDIA_SERVICE.CSFD, get_csfd_tips())
        response = self.api_request(HTTP_METHOD.GET, url)
        self.render_media(handle, response)

    def get_streams(self, handle, url, media_id, root_parent_id, force=False):
        response = self.api_request(HTTP_METHOD.GET, url)
        if len(response) == 0:
            DialogRenderer.ok(get_string(LANG.MISSING_STREAM_TITLE), get_string(LANG.MISSING_STREAM_TEXT))
            return

        selected_stream = None
        if settings.as_bool('auto_select_stream') and not force:
            selected_stream = self._choose_stream_auto(response)

        if selected_stream is None:
            selected_stream = self.choose_stream(response)

        if selected_stream:
            self.process_stream(handle, selected_stream, media_id, root_parent_id)
        else:
            router.set_resolved_url(handle)

    def process_stream(self, *args):
        self.permission_map[settings.as_bool(SETTINGS.PROVIDER_LOGGED_IN)](*args)

    @staticmethod
    def clear_path(handle):
        logger.debug('Clearing path')
        router.replace_route(handle, ROUTE.ROOT)

    @staticmethod
    def show_settings(*args):
        show_settings()

    @staticmethod
    def choose_stream(streams):
        return DialogRenderer.choose_video_stream(streams)

    @staticmethod
    def _choose_stream_auto(response):
        preferred_quality = settings[SETTINGS.STREAM_AUTOSELECT_PREFERRED_QUALITY]
        preferred_languages = [settings.get_select(SETTINGS.PREFERRED_LANGUAGE),
                               settings.get_select(SETTINGS.FALLBACK_LANGUAGE)]
        max_bitrate = settings.as_int('auto_select_stream.max_bitrate') if settings.as_bool(
            'auto_select_stream.limit_bitrate') else 0
        avoid_vcodecs = (settings['auto_select_stream.avoid_video_codecs'] or '').split(',')
        preferred_channels = float(settings['auto_select_stream.preferred_channels']) if settings.as_bool(
            'auto_select_stream.preferred_channels_enabled') else None
        return streams.select(response,
                              audio=preferred_languages,
                              max_bitrate=max_bitrate,
                              vquality=preferred_quality,
                              channels=preferred_channels,
                              codec_blacklist=avoid_vcodecs)

    def api_request(self, *args, **kwargs):
        return self.api_response_handler(self._api.request(*args, **kwargs))

    def watched(self, handle):
        results = self.db[DB_TABLE.WATCH_HISTORY].get_all()
        ids = [row[0] for row in results]
        logger.debug('Getting history list')
        api_res = self.api_request(HTTP_METHOD.GET, self._api.FILTER.ids(ids))
        order = {v: i for i, v in enumerate(ids)}
        api_res['data'] = sorted(api_res['data'], key=lambda x: order[x['_id']])
        self.render_media(handle, api_res, context_menu=ROUTE.WATCHED)

    @staticmethod
    def api_response_handler(response):
        try:
            response.raise_for_status()
            return response.json()
        except requests.exceptions.HTTPError as e:
            InfoDialog(get_string(LANG.HTTP_ERROR), icon=InfoDialogType.ERROR).notify()
            logger.error(e)
            return None

    def _check_token(self):
        return self._provider.is_valid_token()

    def _check_vip(self, user_data):
        logger.debug('Checking VIP')
        is_vip = self._provider.is_vip(user_data)
        if not is_vip:
            logger.debug('VIP is not active')
            InfoDialog(get_string(LANG.ACTIVATE_VIP), sound=True).notify()
            vip_string = get_string(LANG.NOT_ACTIVE)
        else:
            logger.debug('VIP is active')
            vip_string = STRINGS.VIP_INFO.format(self._provider.vip_until(user_data),
                                                 get_string(LANG.DAYS),
                                                 self._provider.vip_remains(user_data))
        settings[SETTINGS.VIP_DURATION] = vip_string
        return is_vip

    def vip_remains(self):
        if time_limit_expired(SETTINGS.LAST_VIP_CHECK, GENERAL.VIP_CHECK_INTERVAL):
            settings[SETTINGS.LAST_VIP_CHECK] = datetime.now()
            valid_token, user_data = self._check_token_and_return_user_data()
            if valid_token:
                if self._check_vip(user_data):
                    days_to = self._provider.vip_remains(user_data)
                    if days_to <= GENERAL.VIP_REMAINING_DAYS_WARN:
                        DialogRenderer.ok(get_string(LANG.VIP_REMAINS).format(provider=self._provider),
                                          STRINGS.PAIR_BOLD.format(get_string(LANG.DAYS), str(days_to)))

    def update_ws_token(self):
        token = self._provider.get_token(settings[SETTINGS.PROVIDER_USERNAME], settings[SETTINGS.PROVIDER_PASSWORD])
        settings[SETTINGS.PROVIDER_TOKEN] = token
        return token

    def _check_token_and_return_user_data(self):
        logger.debug('Checking token and returning user data.')
        user_data = self._provider.get_user_data()
        if not self._provider.is_valid_token(user_data):
            logger.debug('Token is not valid. Getting new one and then new user data.')
            return self.update_ws_token(), self._provider.get_user_data()
        logger.debug('Token is valid.')
        return True, user_data

    def _check_account(self):
        logger.debug('Checking account.')
        valid_token, user_data = self._check_token_and_return_user_data()
        if valid_token:
            logger.debug('Provider token is valid')
            return self._check_vip(user_data)
        else:
            self.info_invalid_credentials()
            self.username_check()
        return False

    @staticmethod
    def show_log_path(handle):
        DialogRenderer.ok(get_string(LANG.LOG_FILE_PATH), get_log_file_path())

    @staticmethod
    def upload_log(handle):
        success, data = read_log()
        if success:
            content = clean_log(data)
            success, response = post_log(content)
            if success:
                show_log_dialog(response)
            else:
                DialogRenderer.ok(get_string(LANG.LOG_FILE_PATH), 'Failed to upload log file: ' + response)
        else:
            DialogRenderer.ok(get_string(LANG.LOG_FILE_PATH), 'Failed to get log file: ' + data)

    @staticmethod
    def _has_preferred_subtitles(stream):
        subs = stream.get('subtitles')
        for sub in subs:
            if sub.get('language') == settings.get_select(SETTINGS.SUBTITLES_PROVIDER_PREFERRED_LANGUAGE):
                logger.debug('Found desired subtitles in the stream')
                return True
        return False

    @staticmethod
    def get_subtitles(media):
        if not media.get('services', {}).get('imdb'):
            return

        try: 
            subs = OpenSubtitles.search(media)
        except:
            return
        
        if subs:
            subs_path = OpenSubtitles.download(subs)
            return subs_path

    def play_stream(self, handle, stream, media_id, root_parent_id):
        if self._check_account():

            logger.debug('Trying to play stream.')
            stream_url = self._provider.get_link_for_file_with_id(stream.get('ident'))
            media = self.api_request(HTTP_METHOD.GET, API.URL.media_detail(media_id))
            labels, art = MediaInfoRenderer.merge_info_labels(media)
            title = MediaInfoRenderer.TITLE.mixed(media)
            labels['filenameandpath'] = MediaInfoRenderer.TITLE.subtitles_search(labels)
            url, item, directory = MediaItem(title, url=stream_url, cast=media.get('cast'), art=art,
                                             info_labels=labels).build()

            if not self._has_preferred_subtitles(stream):
                logger.debug('Subtitles are missing. Trying to find some.')
                subs_path = self.get_subtitles(media)
                if subs_path:
                    item.setSubtitles([subs_path])
            logger.debug('Stream URL found. Playing %s' % stream_url)
            self.send_service_message(SERVICE.PLAYER_SERVICE,
                                      SERVICE_EVENT.PLAYBACK_STARTED,
                                      media_id=root_parent_id)
            router.set_resolved_url(handle, item)
        else:
            router.set_resolved_url()

    @staticmethod
    def send_service_message(service_name, service_event, **kwargs):
        logger.debug('Sending service message {}: {}'.format(service_name, service_event))
        kodi_json_request(SERVICE.PLAYER_SERVICE, SERVICE_EVENT.PLAYBACK_STARTED, kwargs)

    def login_play_stream(self, handle, ident, media_id, root_parent_id):
        self.set_provider_credentials()
        self.play_stream(handle, ident, media_id, root_parent_id)

    @staticmethod
    def info_invalid_credentials():
        InfoDialog(get_string(LANG.INCORRECT_PROVIDER_CREDENTIALS), sound=True).notify()

    def username_check(self):
        username = DialogRenderer.keyboard(get_string(LANG.USERNAME))
        salt = None
        if username:
            salt = self._provider.get_salt(username)

            if salt is None:
                self.info_invalid_credentials()
                return self.username_check()
        return username, salt

    def password_check(self, username, salt):
        password = DialogRenderer.keyboard(get_string(LANG.PASSWORD), hidden=True)
        token = None
        if password:
            password = hash_password(password, salt)
            token = self._provider.get_token(username, password)

            if token is None:
                self.info_invalid_credentials()
                return self.password_check(username, salt)
        return password, token

    @staticmethod
    def set_subtitles_credentials(*args):
        username = DialogRenderer.keyboard(get_string(LANG.USERNAME))
        if username:
            password = DialogRenderer.keyboard(get_string(LANG.PASSWORD), hidden=True)
            if password:
                response = OpenSubtitles.login(username, password)
                if '401' not in response.get('status'):
                    settings[SETTINGS.SUBTITLES_PROVIDER_USERNAME] = username
                    settings[SETTINGS.SUBTITLES_PROVIDER_PASSWORD] = password
                else:
                    StreamCinema.info_invalid_credentials()

    def set_provider_credentials(self, *args):
        username, salt = self.username_check()
        if username and salt is not None:
            password, token = self.password_check(username, salt)
            if password and token:
                settings[SETTINGS.PROVIDER_USERNAME] = username
                settings[SETTINGS.PROVIDER_PASSWORD] = password
                settings[SETTINGS.PROVIDER_TOKEN] = token
                settings[SETTINGS.PROVIDER_LOGGED_IN] = True
                logger.debug('Saving credentials')

    def refresh_provider_token(self, *args):
        username = settings[SETTINGS.PROVIDER_USERNAME]
        password = settings[SETTINGS.PROVIDER_PASSWORD]
        salt = self._provider.get_salt(username)

        if salt is None:
            self.info_invalid_credentials()
            return

        token = self._provider.get_token(username, password)
        if token is None:
            self.info_invalid_credentials()
            return

        settings[SETTINGS.PROVIDER_TOKEN] = token
        DialogRenderer.ok(get_string(LANG.INFO), get_string(LANG.TOKEN_REFRESHED))

    def delete_history_item(self, handle, media_id, *args):
        self.db[DB_TABLE.WATCH_HISTORY].delete(media_id=media_id)
        xbmc.executebuiltin('Container.Refresh')

    def delete_history_all(self, *args):
        self.db[DB_TABLE.WATCH_HISTORY].delete()
        InfoDialog(get_string(LANG.HISTORY_DELETED)).notify()

    @staticmethod
    def set_language(title, language_type):
        languages_values = languages.values()
        selected = DialogRenderer.select(title, languages_values)
        if selected is None:
            return
        selected_language = languages_values[selected]
        selected_language_code = [k for k, v in languages.items() if v == selected_language][0]
        settings.set_select(language_type, selected_language_code)

    def intro(self):
        # WELCOME
        DialogRenderer.ok(get_string(LANG.INTRO_TITLE_1), get_string(LANG.INTRO_TEXT_1))
        # LANGUAGE
        DialogRenderer.ok(get_string(LANG.INTRO_TITLE_2), get_string(LANG.INTRO_TEXT_2))
        StreamCinema.set_language(get_string(LANG.CHOOSE_LANGUAGE), SETTINGS.PREFERRED_LANGUAGE)
        StreamCinema.set_language(get_string(LANG.CHOOSE_FALLBACK_LANGUAGE), SETTINGS.FALLBACK_LANGUAGE)
        # SUBTITLES
        yes = DialogRenderer.yesno(get_string(LANG.INTRO_TITLE_3), get_string(LANG.INTRO_TEXT_3))
        if yes:
            DialogRenderer.ok(get_string(LANG.INTRO_TITLE_3), get_string(LANG.INTRO_TEXT_4))
            StreamCinema.set_subtitles_credentials()
            StreamCinema.set_language(get_string(LANG.INTRO_TITLE_4), SETTINGS.SUBTITLES_PROVIDER_PREFERRED_LANGUAGE)
            StreamCinema.set_language(get_string(LANG.INTRO_TITLE_5), SETTINGS.SUBTITLES_PROVIDER_FALLBACK_LANGUAGE)
        # PROVIDER
        DialogRenderer.ok(get_string(LANG.INTRO_TITLE_6), get_string(LANG.INTRO_TEXT_7))
        self.set_provider_credentials()
        # SEARCH INFO
        DialogRenderer.ok(get_string(LANG.INTRO_TITLE_7), get_string(LANG.INTRO_TEXT_8))
        # ENDING
        DialogRenderer.ok(get_string(LANG.INTRO_TITLE_8), get_string(LANG.INTRO_TEXT_9))
